/************************************************************************
*			     grob2bmp.c V1.0				*
*		     HP48 graphics object to BMP converter		*
*		     Copyright (C) 2003 Robert Tiismus			*
*		          Robert.Tiismus@mail.ee			*
*		        Last update:  03. Jun. 2003			*
*									*
*  This program is free software; you can redistribute it and/or modify	*
*  it under the terms of the GNU General Public License as published by	*
*  the Free Software Foundation; either version 2 of the License, or	*
*  (at your option) any later version.					*
*									*
*  This program is distributed in the hope that it will be useful,	*
*  but WITHOUT ANY WARRANTY; without even the implied warranty of	*
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
*  GNU General Public License for more details.				*
*									*
*  You should have received a copy of the GNU General Public License	*
*  along with this program; if not, write to the Free Software		*
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		*
************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "hputils.h"
#include "grob2bmp.h"

int main( int argc, char *argv[] )
{ int i, j, k, m;		// Just some  counter and cache variables.
  int c; unsigned char rc;	// Temporary character values;
  unsigned long obtype;		// Object type code
  unsigned long grob_filesize;	// The input file size
  unsigned long bm_filesize;	// The output file size
  unsigned long bm_offset;	// Offset from the beginning of file to bitmap
  unsigned long bm_size;	// The BMP bitmap size
  unsigned char *buf;		// Image buffer ( unsigned char pointer )
  unsigned long *ulp;		// Image buffer ( unsigned long pointer )
  unsigned short *usp;		// Image buffer ( unsigned short pointer )

  hpFILE	*grob;		// Input file stream.
  FILE		*bmp;		// Output file stream.
  char	*grobfile_name = NULL;
  char  *bmpfile_name = NULL;
  char verstr[] = "HPHP48-X";	// HP48 file version code string.
  char		HPversion;	// Version letter of the file
  unsigned long horiz, vert;	// Horizontal and vertical pixels.
  unsigned long hbytes;		// Horizontal bytes.
  unsigned long hdwords;	// Horizontal dwords.
  unsigned long obsize;		// GROB object size
  unsigned char cm[ 9 ] = "0x";	// Buffer for colormap argument

  union
   { unsigned char uc[ cm_size ];
     unsigned long ul;
   } colormap[ 2 ] = 
  		{ { BLUE_0, GREEN_0, RED_0, 0 },
		{ BLUE_1, GREEN_1, RED_1, 0 } };

  struct stat sb;		// The input file stat structure

  const char infostr[] =
		"grob2bmp v1.0 by Robert Tiismus <Robert.Tiismus@mail.ee>\n";
  fputs( infostr, stderr );

  // ** Parse the Command-line arguments **
  i = 0;
  while( ++i < argc )
   { if( !strncmp( argv[ i ], "-bw", 3 ) && strlen( argv[ i ] ) == 3 )
      { colormap[ 0 ].ul = 0x00FFFFFFl;
	colormap[ 1 ].ul = 0l;
      }
     else
       if( !strncmp( argv[ i ], "-cm", 3 ) && strlen( argv[ i ] ) == 3 )
	{ if( ++i < argc )
	    if( strlen( argv[ i ] ) == 12 )
	     { strncpy( cm + 2, argv[ i ], 6 );
	       colormap[ 0 ].ul = strtoul( cm, NULL, 16 );
	       strncpy( cm + 2, argv[ i ] + 6, 6 );
	       colormap[ 1 ].ul = strtoul( cm, NULL, 16 );
	     }
	}
       else
	{ if( argv[ i ][ 0 ] != '-' )
	   { if( grobfile_name == NULL )
	       grobfile_name = argv[ i ];
	     else
	       if( bmpfile_name == NULL )
		 bmpfile_name = argv[ i ];
	}
      }
   }

  if( grobfile_name == NULL )
   { fputs( "\nUsage: grob2bmp [options] grob_filename [bmp_filename]\n"
	   "-bw\t\t\tBlack and white colormap\n"
	   "-cm RRGGBBrrggbb\tHex values of 0 and 1 pixel colormap\n\n", stderr );
     exit( 1 );
   }

  // ** Get the input file length **
  if( stat( grobfile_name, &sb ) == -1 )
   { perror( grobfile_name );
     exit( 1 );
   }
  grob_filesize = sb.st_size;

  //** Open the input file **
  if( ( grob = hp_fopen( grobfile_name, "rb" ) ) == NULL )
   { perror( grobfile_name );
     exit( 1 );
   }

  //** Open the output file **
  if( bmpfile_name == NULL )
    bmp = stdout;
  else
    if( ( bmp = fopen( bmpfile_name, "wb" ) ) == NULL )
     { perror( bmpfile_name );
       exit( 1 );
     }

  //********************************************************************
  //                   Process the input file
  //********************************************************************

  //** Check for Version string **                                              
  if( ( fgets( verstr, strlen( verstr ) + 1, grob->stream ) == NULL ) ||
      strncmp( verstr, "HPHP48-", 7 ) )
   { fprintf( stderr, "grob2bmp ERROR: Not a HP48 binary file.\n" );
     exit( 1 );
   }

  // Output the HP48 version string
  HPversion = verstr[ 7 ];
  if( ( m = model( HPversion ) ) != UNKNOWN )
    fprintf( stderr, "HP48%sX file version: %c\n\n",
				m == SSX ? "S/S" : "G/G", HPversion );
  else
    fputs( "HP48 file version: UNKNOWN\n\n", stderr );
    
  // Is it a GROB object?
  if( ( obtype = hpget_a( grob ) ) != DOGROB )
   { fprintf( stderr,
	"grob2bmp ERROR: Nothing to convert: %s object contains no graphics\n",
	     obtype_str( obtype ) );
     exit( 1 );
   }

  // ** Read the GROB size
  obsize = hpget_a( grob );

  // ** Get the GROB dimensions
  vert = hpget_a( grob );
  horiz = hpget_a( grob );
  hbytes = ( horiz - 1l >> 3 ) + 1l;
  hdwords = ( horiz - 1l >> 5 ) + 1l;	// Bitmap rows must be aligned to DW

  // ** Check the GROB size against Vpixel x Hpixel
  if( ( vert * hbytes << 1 ) + 15 != obsize )
   { fprintf( stderr, "grob2bmp ERROR: GROB Object header corrupt \n" );
     exit( 1 );
   }

  // ** Check if GROB file size matches the obsize
  if( ( obsize + 5 >> 1 ) + strlen( verstr ) > grob_filesize )
   { fprintf( stderr, "grob2bmp ERROR: GROB premature end of GROB file\n" );
     exit( 1 ); 
   }

  // ** Initialize the BMP header parameters
  bm_filesize = fh_size + ih_size + 2 * cm_size + ( vert * hdwords  << 2 );
  bm_offset = fh_size + ih_size + 2 * cm_size;
  bm_size = vert * hdwords << 2;

  // Allocate the image buffer
  if( ( buf = ( unsigned char *)malloc( bm_offset + bm_size ) ) == NULL )
   { fputs( "grob2bmp ERROR: Not enough memory to allocate buffer\n", stderr );
     exit( 1 );
   }

  // ** Write the headers
  buf[ ID ] = 'B';		// Signature
  buf[ ID + 1 ] = 'M';
  ulp = ( unsigned long * )( buf + FSIZE );
  ulp[ 0 ] = bm_filesize;	// File size
  ulp[ 1 ] = 0l;		// Reserved
  ulp[ 2 ] = bm_offset;		// Bitmap offset from start of file

  ulp = ( unsigned long * )( buf + fh_size + IH_SIZE );
  ulp[ 0 ] = ih_size;		// Infoheader size
  ulp[ 1 ] = horiz;		// Horizontal pixels
  ulp[ 2 ] = vert;		// Vertical pixels
  usp = ( unsigned short * )( buf + fh_size + PLANES );
  usp[ 0 ] = 1;			// Number of planes
  usp[ 1 ] = 1;			// Bits per pixel
  ulp[ 4 ] = 0l;		// Compression None
  ulp[ 5 ] = bm_size;		// Bitmap size
  ulp[ 6 ] = RESO;		// Horizontal resolution
  ulp[ 7 ] = RESO;		// Vertical resolution
  ulp[ 8 ] = 2l;		// Used colors
  ulp[ 9 ] = 2l;		// Important colors

  ulp = ( unsigned long * )( buf + fh_size + ih_size );
  for( i = 0; i < 2; i++ )
    ulp[ i ] = colormap[ i ].ul;

//     buf[ fh_size+ih_size + I_BLUE + i*cm_size ] = colormap[ i ].uc[ I_BLUE ];
//     buf[ fh_size+ih_size + I_GREEN + i*cm_size ] = colormap[ i ].uc[ I_GREEN ];
//     buf[ fh_size+ih_size + I_RED + i * cm_size ] = colormap[ i ].uc[ I_RED ];
//     buf[ fh_size+ih_size + I_FILL + i * cm_size ] = 0;

  // Fill the buffer with bitmap - Note: it is stored upside down
  for( i = 0; i < vert; i++ )
   { for( j = 0; j < hbytes; j++ )
      { c = hpget_b( grob );
	for( k = 0; k < 8; k++ )
         { rc = ( rc << 1 ) & 0xFE;
	   rc = rc | ( c & 1 );
	   c = c >> 1;
	 }
	buf[ bm_offset + j + ( vert - i - 1 ) * ( hdwords << 2 ) ] = rc;
      }
   }

  // Output the buffer to file
  fwrite( buf, 1, bm_offset + bm_size, bmp );

  free( buf );

  fcloseall();

  return 0;
}
