/************************************************************************
*				HPUTILS.C				*
*	    Library of downloaded HP48G object handling utilities.	*
*                    Copyright (C) 2003 Robert Tiismus                  *
*                         Robert.Tiismus@mail.ee                        *
*                       Last update:  03. Jun. 2003                     *
*                                                                       *
*  This program is free software; you can redistribute it and/or modify *
*  it under the terms of the GNU General Public License as published by *
*  the Free Software Foundation; either version 2 of the License, or    *
*  (at your option) any later version.                                  *
*                                                                       *
*  This program is distributed in the hope that it will be useful,      *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of       *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
*  GNU General Public License for more details.                         *
*                                                                       *
*  You should have received a copy of the GNU General Public License    *
*  along with this program; if not, write to the Free Software          *
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
************************************************************************/
#include <stdio.h>
#include <string.h>
#include "hputils.h"

const struct
{ long obtype;
  int ( *action )( hpFILE * );
  char type_str[ 15 ];
} prologues[] = { { DOBINT,    bint,     "Binary Integer" },
                  { DOREAL,    real,     "Real Number"    },
                  { DOEREL,    erel,     "Long Real"      },
                  { DOCMP,     cmp,      "Complex Number" },
                  { DOECMP,    ecmp,     "Long Complex"   },
                  { DOCHAR,    hpget_b,  "Character"      },
                  { DOARRY,    stdskip,  "Array"          },
                  { DOLNKARRY, stdskip,  "Linked Array"   },
                  { DOCSTR,    stdskip,  "Char String"    },
                  { DOHSTR,    stdskip,  "Hex String"     },
                  { DOLIST,    compskip, "List"           },
                  { DORRP,     rrp,      "Directory"      },
                  { DOSYMB,    compskip, "Algebraic"      },
                  { DOEXT,     compskip, "Unit"           },
                  { DOTAG,     tag,      "Tagged"         },
                  { DOGROB,    stdskip,  "Graphic"        },
                  { DOLIB,     stdskip,  "Library"        },
                  { DOBAK,     stdskip,  "Backup"         },
                  { DOEXT0,    stdskip,  "Library Data"   },
                  { DOACPTR,   acptr,    "Access Pointer" },
                  { DOEXT2,    stdskip,  "External2"      },
                  { DOEXT3,    stdskip,  "External3"      },
                  { DOEXT4,    stdskip,  "External4"      },
                  { DOCOL,     compskip, "Program"        },
                  { DOCODE,    stdskip,  "Code"           },
                  { DOIDNT,    id,       "Global Name"    },
                  { DOLAM,     lam,      "Local Name"     },
                  { DOROMP,    xlib,     "XLIB Name"      } };

/** OBTYPE.STR ****************************************************************
*  Returns the string describing the object type.                             *
*  Input:  Prologue of the object                                             *
*  Output: The object type string.                                            *
******************************************************************************/
const char *obtype_str( long object )
{ int lower = 0,
      upper = sizeof( prologues ) /
 sizeof( prologues[ 0 ] );
  int cont = 5,  /* log2( 28 ) */
         i = ( lower + upper - 1 ) >> 1;
  long ob_i;
  static char ptr[ 10 ];

  while( cont-- )
   { ob_i = prologues[ i ].obtype;
     if( object == ob_i )
       return prologues[ i ].type_str;
     else
      { if( object < ob_i )
          upper = i;
        else
          lower = i;
        i = ( lower + upper ) >> 1;
      }
   }
  sprintf( ptr, "PTR %05lX", object );

  return ptr;
}

/** SKIPOB ********************************************************************
*  Skips object in stream ( repositions the stream pointer ).                 *
*  Input:  Object prologue, input stream positioned stright after prologue    *
*          read.                                                              *
******************************************************************************/
int skipob( long object, hpFILE *in )
{ int lower = 0,
      upper = sizeof( prologues ) / sizeof( prologues[ 0 ] );
  int cont = 5,  /* log2( 28 ) */
         i = ( lower + upper - 1 ) >> 1;
  long ob_i;

  while( cont-- )
   { ob_i = prologues[ i ].obtype;
     if( object == ob_i )
       return ( *prologues[ i ].action )( in );
     else
      { if( object < ob_i )
          upper = i;
        else
          lower = i;
        i = ( lower + upper ) >> 1;
      }
   }

  return 0;
}

/******************************************************************************
*                        Object skipping routines.                            *
******************************************************************************/

int bint( hpFILE *in )
{
  if( hpget_a( in ) == ( long )EOF )
    return EOF;
  return 0;
}

int real( hpFILE *in )
{
  if( hpget_w( in ) == NULL )
    return EOF;
  return 0;
}

int erel( hpFILE *in )
{
  hpget_a( in );
  if( hpget_w( in ) == NULL )
    return EOF;
  return 0;
}

int cmp( hpFILE *in )
{
  real( in );
  return real( in );
}

int ecmp( hpFILE *in )
{
  erel( in );
  return erel( in );
}

int rrp( hpFILE *in )
{ long offset;
  char *nam;

  hpget_x( in );
  if( ( offset = hpget_a( in ) ) == EOF )
    return EOF;
  if( !offset )
    return 0;
  if( hpseek( in, offset - 5L, SEEK_CUR ) )
    return EOF;
  if( ( nam = hpget_name( in ) ) == NULL )
    return EOF;
  if( nam[ 0 ] )
    if( hpget_b( in ) == EOF )
      return EOF;
  return skipob( hpget_a( in ), in );
}

int tag( hpFILE *in )
{
 if( hpget_name( in ) == NULL )
   return EOF;
 return skipob( hpget_a( in ), in );
}

int acptr( hpFILE *in )
{
  hpget_a( in );
  if( hpget_a( in ) == ( long )EOF )
    return EOF;
  return 0;
}

int id( hpFILE *in )
{
  if( hpget_name( in ) == NULL )
    return EOF;
  return 0;
}

int lam( hpFILE *in )
{
  if( hpget_name( in ) == NULL )
    return EOF;
  return 0;
}

int xlib( hpFILE *in )
{
  hpget_x( in );
  return hpget_x( in );
}

int stdskip( hpFILE *in )
{
  if( hpseek( in, hpget_a( in ) - 5, SEEK_CUR ) )
    return EOF;
  return 0;
}

int compskip( hpFILE *in )
{ long obtype;

  while( ( obtype = hpget_a( in ) ) != SEMI )
   { if( obtype == ( long )EOF )
       return EOF;
     if( skipob( obtype, in ) == EOF )
       return EOF;
   }
  return 0;
}

/******************************************************************************
*               Routines to ease reading HP48G stream                         *
******************************************************************************/

hpFILE *hp_fopen( const char *name, const char *mode )
{ FILE *f;
  hpFILE *hf;

  if( ( f = fopen( name, mode ) ) == NULL )
    return NULL;

  if( ( hf = ( hpFILE * )malloc( sizeof( hpFILE ) ) ) == NULL )
   { fprintf( stderr, "ERROR: Not enough memory\n" );
     exit( 1 );
   }

  hf->stream = f;
  hf->odd = 1;
  hf->c = 0;
  hf->last_op = N_OP;
  hf->type = RF;
  
  if( strchr( mode, '+' ) != NULL )
    hf->type = RWF;
  else
    if( mode[ 0 ] == 'w' || mode[ 0 ] == 'a' )
      hf->type = WF;

  return hf;
}

int getnib( hpFILE *in )
{
  in->last_op = R_OP;
  if( in->odd = !in->odd )
   { if( in->c == EOF )
       return EOF;
     return ( in->c >> 4 ) & 0xF;
   }
  else
   { if( ( in->c = getc( in->stream ) ) != EOF )
       return in->c & 0xF;
     else
       return EOF;
   }
}

int hpget_b( hpFILE *in )
{ int c1, c2;

  c1 = getnib( in );
  if( ( c2 = getnib( in ) ) == EOF )
    return c2;

  return c1 | ( c2 << 4 );
}

int hpget_x( hpFILE *in )
{ int c1, c2;

  c1 = getnib( in ) | ( getnib( in ) << 4 );
  if( ( c2 = getnib( in ) ) == EOF )
    return EOF;

  return c1 | ( c2 << 8 );
}

long hpget_a( hpFILE *in )
{ int i, c;
  long result = 0;

  for( i = 0; i < 5; i++ )
   { c = getnib( in );
     result = result | ( ( long )c << ( i << 2 ) );
   }

  if( c == EOF )
    return ( long )c;
  return result;
}

hpword *hpget_w( hpFILE *in )
{ int i, c;
  static hpword word;

  word.hw[ 0 ] = word.hw[ 1 ] = 0L;
  for( i = 0; i < 16; i++ )
   { c = getnib( in );
     word.bytes[ i >> 1 ] |= ( unsigned char )c << ( ( i % 2 ) << 2 );
   }

  if( c == EOF )
    return NULL;
  return &word;
}

char *hpget_name( hpFILE *in )
{ static char namebuff[ 256 + 1 ];
  int namelen, i;

  if( ( namelen = hpget_b( in ) ) == EOF )
    return NULL;
  namebuff[ 0 ] = 0;
  if( namelen )
   { for( i = 0 ; i < namelen; i++ )
       if( ( namebuff[ i ] = hpget_b( in ) ) == EOF )
         return NULL;
     namebuff[ namelen ] = 0;
   }

  return namebuff;
}

unsigned do_crc( hpFILE *in, long count )
{ unsigned crc = 0;
  long i;

  for( i = 0; i < count; i++ )
    crc = ( crc >> 4 ) ^ ( ( ( crc & 0xF ) ^ getnib( in ) ) * 0x1081 );

  return crc;
}

/******************************************************************************
*               Routines to ease writing HP48G stream                         *
******************************************************************************/

void putnib( int nib, hpFILE *out )
{
  out->last_op = W_OP;
  if( out->odd = !out->odd )
   { out->c = ( out->c & 0xF ) | ( ( nib & 0xF ) << 4 );
     putc( out->c, out->stream );
   }
  else
    out->c = nib & 0xF;
}

void hpput_b( int byte, hpFILE *out )
{
  putnib( byte, out );
  putnib( byte >> 4, out );
}

void hpput_x( int ext, hpFILE *out )
{ int i;

  for( i = 0; i < 3; i++ )
    putnib( ext >> ( i << 2 ), out );
}

void hpput_a( long addr, hpFILE *out )
{ int i;

  for( i = 0; i < 5; i++ )
    putnib( ( int )( addr >> ( i << 2 ) ), out );
}

void hpput_w( hpword word, hpFILE *out )
{ int i;

  for( i = 0; i < 16; i++ )
    putnib( ( int )( word.bytes[ i >> 2 ]  >> ( ( i % 2 ) << 2 ) ), out );
}

int hp_fclose( hpFILE *out )
{ int result;

  if( out->last_op == W_OP )
    switch( out->type )
     { case WF:
         if( !out->odd )
           putnib( 0, out );
         break;
       case RWF:
         if( !out->odd )
           hpseek( out, 0, SEEK_CUR );
     }

  result = fclose( out->stream );
  free( out );
  
  return result;
}

/******************************************************************************
*               Routines to ease handling HP48G stream                        *
******************************************************************************/

int hpseek( hpFILE *in, long offset, int whence )
{ long local_offs;
  int result, c;

  if( in->last_op == W_OP )
   { switch( in->type )
       { case WF:
           if( !in->odd )
            { putnib( 0, in );
	      fseek( in->stream, -1, SEEK_CUR );
	      in->odd = !in->odd;
	    }
           break;
         case RWF:
           if( !in->odd )
            { fseek( in->stream, 0, SEEK_CUR );
              c = ( getc( in->stream ) >> 4 ) & 0xF;
              fseek( in->stream, -1, SEEK_CUR );
              putnib( c, in );
              fseek( in->stream, -1, SEEK_CUR );
	      in->odd = !in->odd;
            }
       }
   }

  in->last_op = N_OP;

  switch( whence )
   { case SEEK_CUR:
       local_offs = offset / 2;
       if( !in->odd && !( offset % 2 ) )
         --local_offs;
       result = fseek( in->stream, local_offs, SEEK_CUR );
       in->odd = in->odd ^ ( offset % 2 );
       break;
     case SEEK_SET:
       local_offs = offset / 2;
       result = fseek( in->stream, local_offs, SEEK_SET );
       in->odd = !( offset % 2 );
   }

  if( !( in->odd || result ) )
   { in->odd = !in->odd;
     if( getnib( in ) == EOF )
       result = -1;
   }

  return result;
}

void hpsync_w( hpFILE *hf )
{
  if( hf->last_op == R_OP )
    fseek( hf->stream, -1, SEEK_CUR );
}

long hptell( hpFILE *in )
{ long result;

  if( ( result = ftell( in->stream ) ) != -1L )
   { result *= 2;
     if( !in->odd )
       --result;
   }

  return result;
}

// **************************************************************************
// * Discover the Calculator model Model from whitch the file was downloaded
// **************************************************************************
int model( char ver )
{ const char HP48S[ 8 ] = "ABCDEFJ";
  const char HP48G[ 7 ] = "KLMNPR";

  if( strchr( HP48S, ver ) != NULL )
   return SSX;
  else
   { if( strchr( HP48G, ver ) != NULL )
      return GGX;
     else
      return UNKNOWN;
   }
}
